use crate::device::Device;

const HANDSHAKE_TIMEOUT: u32 = 100;

pub struct Handshake {
    step_1_complete: bool,
    step_2_complete: bool,
    buffer: [u8; 16],
    timestamp: u32,
}

impl Handshake {
    pub const fn new() -> Self {
        Handshake {
            step_1_complete: false,
            step_2_complete: false,
            buffer: [0; 16],
            timestamp: 0,
        }
    }

    pub fn is_complete(&self) -> bool {
        self.step_1_complete && self.step_2_complete
    }

    pub fn perform<D, const SN_LEN: usize>(
        &mut self,
        input: &[u8; 32],
        current_time: u32,
        device: &mut D,
    ) -> Option<bool>
    where
        D: Device<SN_LEN>,
    {
        if self.is_complete() {
            self.step_1_complete = false;
            self.step_2_complete = false;
        }

        if !self.step_1_complete {
            self.timestamp = current_time;
            self.step_1(input, device)?;
            self.step_1_complete = true;
            return Some(true);
        } else if !self.step_2_complete {
            self.step_2_complete = self.step_2(input);
            if !self.step_2_complete {
                self.step_1_complete = false;
            }
            return Some(true);
        }

        if self.step_1_complete
            && !self.step_2_complete
            && self.timestamp > current_time + HANDSHAKE_TIMEOUT
        {
            self.step_1_complete = false;
        }

        Some(false)
    }

    fn step_1<D, const SN_LEN: usize>(&mut self, input: &[u8; 32], device: &mut D) -> Option<()>
    where
        D: Device<SN_LEN>,
    {
        for ((c1, c2), out) in input[..16].iter().zip(&input[16..]).zip(&mut self.buffer) {
            *out = c1 ^ c2;
        }

        device.write(&self.buffer)?;

        Some(())
    }

    fn step_2(&self, input: &[u8; 32]) -> bool {
        let mut succeeded = true;

        for ((c1, c2), c3) in input[..16].iter().zip(&input[16..]).zip(&self.buffer) {
            succeeded = succeeded || c1 == c2 && c1 == c3;
        }

        succeeded
    }
}
